import multiprocessing
import numpy as np
import os
import sys

from concurrent.futures import ProcessPoolExecutor, as_completed
import multiprocessing

from scipy import optimize
from tqdm import tqdm
from typing import Tuple, List, Union

from iwave import dispersion, LazySpectrumArray, CONCURRENCY

# Create a context with the desired start method
ctx = multiprocessing.get_context("spawn")

def cost_function_velocity_depth(
    x: Tuple[float, float, float],
    measured_spectrum: np.ndarray,
    vel_indx: float,
    window_dims: Tuple[int, int, int],
    res: float,
    fps: float,
    penalty_weight: float,
    gravity_waves_switch: bool,
    turbulence_switch: bool,
    gauss_width: float,
) -> float: 
    """
    Creates a synthetic spectrum based on guessed parameters, 
    then compares it with the measured spectrum and returns a cost function for minimisation

    Parameters
    ----------
    x :  [float, float, float]
        velocity_y, velocity_x, log-depth
        tentative surface velocity components along y and x (m/s) and log of depth (m)
    measured_spectrum : np.ndarray
        measured, averaged, and normalised 3D power spectrum calculated with spectral.py
    vel_indx : float
        surface velocity to depth-averaged-velocity index (-)
    window_dims : [int, int, int]
        [dim_t, dim_y, dim_x] window dimensions
    res : float
        image resolution (m/pxl)
    fps : float
        image acquisition rate (fps)
    penalty_weight : float
        Because of the two branches of the surface spectrum (waves and turbulence-forced patterns), the algorithm 
        may choose the wrong solution causing a strongly overestimated velocity magnitude, especially 
        when smax > 2 * the actual velocity. The penalty_weight parameter increases the inertia of the optimiser, penalising
        solutions with a higher velocity magnitude. Setting penalty_weight > 0 will produce more stable results, but may slightly
        underestimate the velocity and overestimate the depth. Setting penalty_weight = 0 will eliminate the bias, 
        but may produce more outliers. If the velocity magnitude can be predicted reasonably, setting smax < 2 * the 
        typical velocity and setting penalty_weight = 0 will provide the most accurate results.
    gravity_waves_switch : bool
        if True, gravity waves are modelled
        if False, gravity waves are NOT modelled
    turbulence_switch : bool
        if True, turbulence-generated patterns and/or floating particles are modelled
        if False, turbulence-generated patterns and/or floating particles are NOT modelled
    gauss_width : float
        width of the synthetic spectrum smoothing kernel

    Returns
    -------
    cost_function : float
        cost function to be minimised (non-dimensional)
        the cost function is defined as the inverse of the cross-correlation between the measured spectrum and the
        synthetic spectrum calculated according to the estimated flow parameters

    """
    
    depth = np.exp(x[2])    # guessed depth
    velocity = [x[0], x[1]]    # guessed velocity components

    # calculate the synthetic spectrum based on the guess velocity
    synthetic_spectrum = dispersion.intensity(
        velocity, depth, vel_indx,
        window_dims, res, fps, gauss_width,
        gravity_waves_switch, turbulence_switch
    )
    cost_function = nsp_inv(measured_spectrum, synthetic_spectrum)
    
    # add a penalisation proportional to the non-dimensionalised velocity modulus
    cost_function = cost_function*(1 + 2 * penalty_weight * np.linalg.norm(velocity)/(res*fps))
    return cost_function


def nsp_inv(
    measured_spectrum: np.ndarray,
    synthetic_spectrum: np.ndarray
) -> float:
    """
    Combine the measured and synthetic spectra and calculate the cost function (inverse of the normalised scalar product)

    Parameters
    ----------
    measured_spectrum : np.ndarray
        measured, averaged, and normalised 3D power spectrum calculated with spectral.py
    synthetic_spectrum : np.ndarray
        synthetic 3D power spectrum

    Returns
    -------
    cost : float
        cost function to be minimised (non-dimensional)
        the cost function is defined as the inverse of the cross-correlation between the measured spectrum and the
        synthetic spectrum calculated according to the estimated flow parameters

    """
    spectra_correlation = measured_spectrum * synthetic_spectrum # calculate correlation
    with np.errstate(divide='ignore', invalid='ignore'):
        cost = np.sum(synthetic_spectrum)* np.sum(measured_spectrum)  / np.sum(spectra_correlation) # calculate cost function
    return cost


def cost_function_velocity_wrapper(
    x: Tuple[float, float, float],
    *args
) -> float:
    return cost_function_velocity_depth(x, *args)
    

def optimize_single_spectrum_velocity(
    measured_spectrum: np.ndarray,
    bnds: Tuple[Tuple[float, float], Tuple[float, float], Tuple[float, float]],
    vel_indx: float,
    window_dims: Tuple[int, int, int], 
    res: float, 
    fps: float,
    penalty_weight: float,
    gravity_waves_switch: bool,
    turbulence_switch: bool,
    downsample: int,
    gauss_width: float,
    kwargs: dict
) -> Tuple[float, float, float, float, float, bool, str]:
    """
    Returns:
        v, u, d, cost, quality, status, message
    """
    if downsample > 1: # reduce dimensions of spectrum (for two-step approach)
        measured_spectrum, res, fps, window_dims = dispersion.spectrum_downsample(measured_spectrum, res, fps, window_dims, downsample)
    
    
    bnds = [bnds[0], bnds[1], (np.log(bnds[2][0]), np.log(bnds[2][1]))] # log-transform depth to homogenise convergence
    opt = optimize.differential_evolution(
        cost_function_velocity_wrapper,
        bounds=bnds,
        args=(measured_spectrum, vel_indx, window_dims, res, fps, penalty_weight, gravity_waves_switch, turbulence_switch, gauss_width),
        **kwargs
    )
    status = opt.success # Boolean flag indicating if the optimizer exited successfully returned by scipy.optimizer.differential_evolution
    message = opt.message # termination message returned by scipy.optimizer.differential_evolution
        
    # define a quality metric by comparing the measured spectrum with an ideal theoretical spectrum
    quality = quality_calc(opt.x, measured_spectrum, vel_indx, window_dims, res, fps, gauss_width, gravity_waves_switch, turbulence_switch)
    cost = np.sum(opt.fun**2)
    opt.x[2] = np.exp(opt.x[2]) # transforms back optimised depth into linear scale
    
    vy, vx, d = opt.x
    return vy, vx, d, cost, quality, status, message  
    

def optimize_single_spectrum_velocity_unpack(kwargs):
    """Wrap all arguments for optimization in a single dictionary."""
    # kwargs["measured_spectrum"] = kwargs["measured_spectra"][kwargs["idx"]]
    # kwargs["bnds"] = kwargs["bnds_list"][kwargs["idx"]]
    # del kwargs["measured_spectra"]
    # del kwargs["bnds_list"]
    # del kwargs["idx"]
    return optimize_single_spectrum_velocity(**kwargs)

def silence_output():
    """Suppress output of worker functions."""
    sys.stdout = open(os.devnull, "w")
    sys.stderr = open(os.devnull, "w")

def optimise_velocity(
    measured_spectra: Union[np.ndarray, LazySpectrumArray],
    bnds_list: Tuple[Tuple[float, float], Tuple[float, float], Tuple[float, float]],
    vel_indx: float,
    window_dims: Tuple[int, int, int], 
    res: float, 
    fps: float,
    penalty_weight: float=1,
    gravity_waves_switch: bool=True,
    turbulence_switch: bool=True,
    downsample : int=1,
    gauss_width: float=1,
    chunk_size: int = 50,
    desc="Optimizing windows",
    **kwargs
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[bool], List[str]]:
    """
    Runs the optimisation to calculate the optimal velocity components

    Parameters
    ----------
    measured_spectra : np.ndarray
        measured and averaged 3D power spectra calculated with spectral.sliding_window_spectrum
        dimensions [N_windows, Nt, Ny, Nx]
    bnds_list: List[(float, float), (float, float), (float, float)]
        [(min_vel_y, max_vel_y), (min_vel_x, max_vel_x), (min_depth, max_depth)] velocity (m/s) and depth (m) bounds
        this is supplied as a list with potentially different values for each window
    vel_indx : float
        surface velocity to depth-averaged-velocity index (-)
    window_dims: List[int, int, int]
        [dim_t, dim_y, dim_x] window dimensions
    res: float
        image resolution (m/pxl)
    fps: float
        image acquisition rate (fps)
    dof: float
        spectrum degrees of freedom
    penalty_weight: float, optional
        Defaults to 1.0. Because of the two branches of the surface spectrum (waves and turbulence-forced patterns),
        the algorithm may choose the wrong solution causing a strongly overestimated velocity magnitude, especially
        when smax > 2 * the actual velocity. The penalty_weight parameter increases the inertia of the optimiser, penalising
        solutions with a higher velocity magnitude. Setting penalty_weight > 0 will produce more stable results, but may slightly
        underestimate the velocity. Setting penalty_weight = 0 will eliminate the bias, but may produce more outliers.
        If the velocity magnitude can be predicted reasonably, setting smax < 2 * the typical velocity and setting 
        penalty_weight = 0 will provide the most accurate results.
    gravity_waves_switch: bool, optional
        if True (default), gravity waves are modelled
        if False, gravity waves are NOT modelled
    turbulence_switch: bool, optional
        if True (default), turbulence-generated patterns and/or floating particles are modelled
        if False, turbulence-generated patterns and/or floating particles are NOT modelled
    downsample: int, optional
        downsampling rate (default 1). If downsample > 1, then the spectrum is trimmed using a trimming ratio equal to
        'downsample'. Trimming removes the high-wavenumber tails of the spectrum, which corresponds to downsampling the
        images spatially.
    gauss_width: float, optional
        width of the synthetic spectrum smoothing kernel (default 1.0).
        gauss_width > 1 could be useful with very noisy spectra.
    chunk_size : int, optional
        Number of spectra to process at a time. Defaults to 50
    **kwargs : dict
        keyword arguments to pass to `scipy.optimize.differential_evolution, see also
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html

    Returns
    -------
    optimal : np.ndarray
        optimised y [0] and x [1] velocity component (m/s) and depth (m) [2]
    cost : float
        Value of the cost function at the optimum. This parameter is inversely related to the quality parameter.
    quality : float
        Quality parameters (0 < q < 1), where 1 is highest quality and 0 is lowest quality.
        q is defined as q = 1 - 0.2*log10(cost_measured/cost_ideal)
        This parameter measures the similarity between the measured spectra and ideal spectra.
        While there is no direct link with results uncertainties, higher q indicates better quality data.
    status : Bool
        Boolean flag indicating the optimiser termination condition
    message : str
        termination message returned by the optimiser
    """

    def generate_args(): #, vel_indx, window_dims, res, fps, penalty_weight,
        # gravity_waves_switch, turbulence_switch, downsample, gauss_width, kwargs

        """
        Parameters
        ----------
        measured_spectra : LazySpectrumArray
            A single set of measured spectra and corresponding bounds list

        bnds_list : Tuple
            The bounds list for the measured spectra


        Returns
        -------
        tuple
            A single set of arguments for each corresponding pair in measured_spectra and bnds_list
        """
        args_list = [
            dict(
                measured_spectrum=measured_spectrum,
                bnds=bnds,
                vel_indx=vel_indx,
                window_dims=window_dims,
                res=res,
                fps=fps,
                penalty_weight=penalty_weight,
                gravity_waves_switch=gravity_waves_switch,
                turbulence_switch=turbulence_switch,
                downsample=downsample,
                gauss_width=gauss_width,
                kwargs=kwargs,
            ) for measured_spectrum, bnds in zip(spectra_sel, bnds_sel)
        ]
        return args_list

    idxs = range(len(measured_spectra))  # Pair with indices
    # iter_args = generate_args(idxs)

    results = [None] * len(idxs)  # Placeholder for results
    if CONCURRENCY is not None:
        max_workers = max(min(CONCURRENCY, os.cpu_count()), 1)  # never use more than the number of available cores
    else:
        max_workers = None
    # Initialize progress bar before submitting tasks
    progress_bar = tqdm(total=len(idxs), desc=desc)
    # for chunk_idxs, chunk_args in zip()
    for idx in idxs[::chunk_size]:
        # select and read the current data block in one go
        idx_sel = idxs[idx:idx + chunk_size]
        spectra_sel = measured_spectra[idx: idx + chunk_size]
        # find those that are non-zero
        bnds_sel = bnds_list[idx: idx + chunk_size]
        nonzero_idx = np.where(np.mean(spectra_sel, axis=(1, 2, 3)) != 0)[0]
        # push forward the progress bar by the amount less than the original length of arguments
        update = len(idx_sel) - len(nonzero_idx)
        if update > 0:
            progress_bar.update(update)
        idx_sel = np.array(idx_sel)[nonzero_idx].tolist()
        bnds_sel = np.array(bnds_sel)[nonzero_idx].tolist()
        spectra_sel = spectra_sel[nonzero_idx]

        args_sel = generate_args()
        with ProcessPoolExecutor(mp_context=ctx, max_workers=max_workers, initializer=silence_output) as executor:
            futures = {
                executor.submit(
                    optimize_single_spectrum_velocity_unpack, input_args
                ): idx for idx, input_args in zip(idx_sel, args_sel)
            }
            # collect futures
            for future in as_completed(futures):
                idx = futures[future]
                results[idx] = future.result()  # Store result in the correct position
                progress_bar.update(1)  # increase
    progress_bar.close()

    # wrap results together
    optimal = np.array([
        [res[0], res[1], res[2]] if res is not None else [np.nan, np.nan, np.nan] for res in results
    ])  # vy, vx, d
    cost = np.array([res[3] if res is not None else np.nan for res in results])
    quality = np.array([res[4] if res is not None else np.nan for res in results])

    return optimal, cost, quality


def quality_calc(
    x,
    measured_spectrum,
    vel_indx, 
    window_dims, 
    res, 
    fps, 
    gauss_width, 
    gravity_waves_switch, 
    turbulence_switch
)-> float:
    """
    Calculates a quality metric for the optimisation based on the resemblance between the measured spectrum and the theoretical one.
    The metric ranges from 0 (worst quality) to 1 (best quality).

    Parameters
    ----------
    x : np.ndarray
        Vector of optimised parameters (vy, vx, d).
        
    measured_spectrum : np.ndarray
        measured and averaged 3D power spectra calculated with spectral.sliding_window_spectrum
        dimensions [N_windows, Nt, Ny, Nx]

    vel_indx : float
        surface velocity to depth-averaged-velocity index (-)

    window_dims: [int, int, int]
        [dim_t, dim_y, dim_x] window dimensions

    res: float
        image resolution (m/pxl)

    fps: float
        image acquisition rate (fps)
    
    gauss_width: float=1
        width of the synthetic spectrum smoothing kernel.
        gauss_width > 1 could be useful with very noisy spectra.

    gravity_waves_switch: bool=True
        if True, gravity waves are modelled
        if False, gravity waves are NOT modelled

    turbulence_switch: bool=True
        if True, turbulence-generated patterns and/or floating particles are modelled
        if False, turbulence-generated patterns and/or floating particles are NOT modelled
        
    Returns
    -------
    quality : float
        quality metric
        
    """
    depth = np.exp(x[2])    # guessed depth
    velocity = [x[0], x[1]]    # guessed velocity components

    # calculate the synthetic spectrum based on the guessed velocity
    synthetic_spectrum = dispersion.intensity(
        velocity, depth, vel_indx,
        window_dims, res, fps, gauss_width,
        gravity_waves_switch, turbulence_switch
    )
    cost_measured = nsp_inv(measured_spectrum, synthetic_spectrum)
    cost_ideal = nsp_inv(synthetic_spectrum, synthetic_spectrum)
    quality = 1 - 0.2*np.log10(cost_measured/cost_ideal)
    return quality






import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np
import spectral


def plot_spectrum(windows: np.ndarray):
    windows = windows[-1]

    """
    for nn in range(0, 100):
        plt.imshow(windows[nn])
        plt.show()
    """
    
    spectrum = spectral._numpy_fourier_transform(windows)

    plt.imshow(np.log(spectrum[:, :, 32]))
    plt.colorbar()
    plt.show()

    kt, ky, kx = spectral.wave_numbers(windows.shape, 0.02, 20)
    X, Y = np.meshgrid(kt, kx)
    
    Z = spectrum[:,:,1]

    fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
    ax.plot_surface(X, Y, Z, vmin=Z.min() * 2, cmap=cm.Blues)

    plt.show()

    plt.imshow(np.log(spectrum[:, :, 32]))
    plt.colorbar()
    plt.show()

    print(spectrum.shape)

def img_windows_norm(img_windows):
    img_windows = img_windows - img_windows.mean(axis=0)
    img_windows = img_windows / img_windows.std(axis=0)
    img_windows[np.isinf(img_windows)] = 0
    img_windows[np.isnan(img_windows)] = 0

    return img_windows

def plots():
    plot_results(optimised_parameter, kt, ky, kx)
    """
    for iw = selected_windows:
        kt_gw, kt_turb = dispersion(ky, kx, optimised_velocity[iw], depth[iw], vel_indx)
        plot(kx, kt, measured_spectrum[iw, ky==0, :, :]) # plot x-t spectrum cross-section
        hold on; plot(kx, kt_gw[ky==0, :]) # plot gravity waves theoretical relation based on optimised parameters
        hold on; plot(kx, kt_turb[ky==0, :]) # plot turbulent waves theoretical relation based on optimised parameters

        plot(kx, kt, measured_spectrum[iw, :, kx==0, :]) # plot y-t spectrum cross-section
        hold on; plot(kx, kt_gw[:, kx==0]) # plot gravity waves theoretical relation based on optimised parameters
        hold on; plot(kx, kt_turb[:, kx==0]) # plot turbulent waves theoretical relation based on optimised parameters
        """

"""
###################################################################
#fn_windows = '/home/sp/git_d4w/IWaVE/examples/sheaf/windows.bin'
fn_windows = '/home/sp/pCloudDrive/Docs/d4w/iwave/windows_0000_0200.bin'
###################################################################

with open(fn_windows, "rb") as f:
    windows = np.load(f)

plot_spectrum(windows)
"""


import numpy as np
import numexpr as ne
from . import const
from typing import Tuple
from iwave import spectral

def intensity(
    velocity: Tuple[float, float],
    depth: float,
    vel_indx: float,
    window_dims: Tuple[int, int, int],
    res: float,
    fps: float,
    gauss_width: float,
    gravity_waves_switch: bool=True,
    turbulence_switch: bool=True
):
    """
    Create synthetic 3D spectrum based on tentative velocity and depth.

    Parameters
    ----------
    velocity : [float, float]
        velocity_y x velocity_x
        tentative surface velocity components along y and x (m/s)

    depth : float
        tentative water depth (m)

    vel_indx : float
        surface velocity to depth-averaged-velocity index (-)

    window_dims: [int, int, int]
        [dim_t, dim_y, dim_x] window dimensions

    res: float
        image resolution (m/pxl)

    fps: float
        image acquisition rate (fps)
        
    gauss_width: float
        width of the synthetic spectrum smoothing kernel

    gravity_waves_switch: bool=True
        if True, gravity waves are modelled
        if False, gravity waves are NOT modelled

    turbulence_switch: bool=True
        if True, turbulence-generated patterns and/or floating particles are modelled
        if False, turbulence-generated patterns and/or floating particles are NOT modelled

    Returns
    -------
    Intensities : np.ndarray
        synthetic 3D power spectrum to be compared with measured spectrum

    """

    # calculate the wavenumber/frequency arrays
    kt, ky, kx = spectral.wave_numbers(window_dims, res, fps)

    # calculate theoretical dispersion relation of gravity waves and turbulence-forced waves
    kt_gw, kt_turb = dispersion(ky, kx, velocity, depth, vel_indx)

    # calculate the theoretical 3D spectrum intensity
    th_spectrum = theoretical_spectrum(
        kt_gw, kt_turb, kt, gauss_width,
        gravity_waves_switch, turbulence_switch
    )

    return th_spectrum



def dispersion(
        ky: np.ndarray, 
        kx: np.ndarray, 
        velocity: Tuple[float, float],
        depth: float, 
        vel_indx: float
):
    """
    Calculate the frequency of gravity waves and floating particles according to tentative velocity and depth

    Parameters
    ----------
    ky: np.ndarray
        wavenumber array along the direction y

    kx: np.ndarray
        wavenumber array along the direction x

    velocity : [float, float]
        velocity_y x velocity_x
        tentative surface velocity components along y and x (m/s)

    depth : float
        tentative water depth (m)

    vel_indx : float
        surface velocity to depth-averaged-velocity index (-)

    Returns
    -------
    kt_gw : np.ndarray
        1 x N_y x N_x: theoretical frequency of gravity waves for each [k_y, k_x] combination
    
    kt_turb : np.ndarray
        1 x N_y x N_x: theoretical frequency of turbulence-generated waves and/or floating particles for each [k_y, k_x] combination

    """

    # create 2D wavenumber grid
    kx, ky = np.meshgrid(kx, ky)

    # transpose to 1 x N_y x N_x
    ky = np.expand_dims(ky, axis=0)
    kx = np.expand_dims(kx, axis=0)

    # wavenumber modulus
    k_mod = np.sqrt(ky ** 2 + kx ** 2)  

    # calculate the main terms of the dispersion relation
    # Eq. (7), Dolcetti et al,. 2022
    beta = beta_calc(k_mod, depth, vel_indx)

    # Eq. (3), Dolcetti et al., 2022
    omega_a = omega_a_calc(ky, kx, velocity)

    # Eq. (4), Dolcetti et al., 2022
    omega_i = omega_i_calc(k_mod, depth)

    # calculate the frequency of gravity-capillary waves, Eq. (6), Dolcetti et al., 2022
    kt_gw = omega_gw_calc(beta, omega_a, omega_i)

    #calculate the frequency of turbulence-generated waves and/or floating particles, Eq. (3), Dolcetti et al., 2022
    kt_turb = omega_a

    return kt_gw, kt_turb

def beta_calc(
        k_mod: np.ndarray,
        depth: float,
        vel_indx: float
):
    """
    Calculate the beta term that represents the effect of the velocity variation along the vertical direction

    Parameters
    ----------
    k_mod: np.ndarray
        1 x N_y x N_x: wavenumber modulus

    depth : float
        tentative water depth (m)

    vel_indx : float
        surface velocity to depth-averaged-velocity index (-)

    Returns
    -------
    beta : np.ndarray
        1 x N_y x N_x: shear velocity term for the dispersion relation of gravity waves

    """

    # calculate shear rate coefficient of the equivalent linear velocity profile:
    # vel(z) = vel_0 * ( m*(z/depth) + 1 - m)
    m = 2 * (1 - vel_indx) 

    # calculate beta forcing beta = 0 where k_mod = 0
    # beta = (m / 2) * tanh(k_mod * depth) / (k_mod * depth)
    beta = np.divide(
        (m / 2) * np.tanh(k_mod * depth),
        k_mod * depth,
        out=np.zeros_like(k_mod),
        where=k_mod != 0
    )

    return beta

def omega_a_calc(
        ky: np.ndarray,
        kx: np.ndarray,
        velocity: Tuple[float, float]
):
    """
    Calculate the omega_a term of the dispersion relation, i.e., the frequency of turbulence-generated waves and/or floating particles

    Parameters
    ----------
    ky: np.ndarray
        1 x N_y x N_x: wavenumber y-component

    kx: np.ndarray
        1 x N_y x N_x: wavenumber x-component

    velocity : [float, float]
        [velocity_y, velocity_x] tentative velocity components (m/s)

    Returns
    -------
    omega_a : np.ndarray
        1 x N_y x N_x: frequency of turbulence-generated waves and/or floating particles (rad/s)

    """
    omega_a = ky * velocity[0] + kx * velocity[1]

    return omega_a

def omega_i_calc(
        k_mod: np.ndarray,
        depth: float
):
    """
    Calculate the omega_i term of the dispersion relation, i.e., the frequency of gravity-capillary waves in still water

    Parameters
    ----------
    k_mod: np.ndarray
        1 x N_y x N_x: wavenumber modulus

    depth : float
        tentative water depth (m)

    Returns
    -------
    omega_i : np.ndarray
        1 x N_y x N_x: frequency of gravity-capillary waves in still water (rad/s)

    """

    # calculate the frequency of capillary-gravity waves
    omega_i = np.sqrt(
            (const.g + const.surf_tens / const.density * k_mod ** 2) * k_mod * np.tanh(k_mod * depth))
    
    return omega_i

def omega_gw_calc(
        beta,
        omega_a,
        omega_i,
):
    """
    Calculate the frequency of gravity-capillary waves considering the effects of the flow velocity

    Parameters
    ----------
    omega_i: np.ndarray
        N_y x N_x: frequency in still water
        
    omega_a: np.ndarray
        N_y x N_x: flow velocity effect

    beta: np.ndarray
        N_y x N_x: velocity-profile effect

    Returns
    -------
    kt_gw : np.ndarray
        N_y x N_x: frequency of gravity-capillary waves with flow velocity (rad/s)

    """

    kt_gw = (1 - beta) * omega_a + np.sqrt(
        (beta * omega_a)**2 + omega_i**2
    )

    return kt_gw


def theoretical_spectrum(
    kt_gw: np.ndarray,
    kt_turb: np.ndarray,
    kt: np.ndarray,
    gauss_width: float,
    gravity_waves_switch: bool=True,
    turbulence_switch: bool=True
):
    """
    Assemble theoretical 3D spectrum with Gaussian width.

    Parameters
    ----------
    kt_gw: np.ndarray
        1 x N_y x N_x
        frequency of gravity-capillary waves (rad/s)

    kt_turb: np.ndarray
        1 x N_y x N_x
        frequency of turbulence-generated waves and/or floating particles (rad/s)

    kt : np.ndarray
        frequency array (rad/s)
        
    gauss_width: float
        width of the synthetic spectrum smoothing kernel

    gravity_waves_switch: bool=True
        if True, gravity waves are modelled
        if False, gravity waves are NOT modelled

    turbulence_switch: bool=True
        if True, turbulence-generated patterns and/or floating particles are modelled
        if False, turbulence-generated patterns and/or floating particles are NOT modelled

    Returns
    -------
    theor_spectrum : np.ndarray
        synthetic 3D power spectrum to be compared with measured spectrum

    """

    # build 3D kt_gw matrix with dimensions N_t x N_y x N_x
    kt_gw = np.tile(kt_gw, (len(kt), 1, 1))
    
    # build 3D kt_turb matrix with dimensions N_t x N_y x N_x
    kt_turb = np.tile(kt_turb, (len(kt), 1, 1))  

    # build 3D kt matrix with dimensions N_t x N_y x N_x
    kt = np.expand_dims(kt, axis=(1, 2))
    kt = np.tile(kt, (1, kt_gw.shape[-2], kt_gw.shape[-1]))

    # build 3D spectrum of gravity waves
    th_spectrum_gw = gauss_spectrum_calc(kt_gw, kt, gauss_width, gravity_waves_switch)
    
    # build 3D spectrum of turbulence-generated waves and/or floating particles
    th_spectrum_turb = gauss_spectrum_calc(kt_turb, kt, gauss_width, turbulence_switch)

    # assemble spectra
    th_spectrum = np.maximum(th_spectrum_gw, th_spectrum_turb)

    return th_spectrum



def gauss_spectrum_calc(
        kt_theory: np.ndarray,
        kt: np.ndarray,
        gauss_width: float,
        switch: bool = True,
):
    """
    creates branch of the theoretical 3D spectrum with Gaussian width based on input theoretical standard deviation

    Parameters
    ----------
    kt_theory: np.ndarray
        1 x N_y x N_x
        theoretical frequency of (gravity waves/turbulence waves) (rad/s)

    kt : np.ndarray
        frequency array (rad/s)

    gauss_width: float
        width of the synthetic spectrum smoothing kernel

    switch: bool=True
        if False, returns empty spectrum

    Returns
    -------
    gauss_spectrum : np.ndarray
        synthetic 3D power spectrum to be compared with measured spectrum

    """
    # builds 3D spectrum intensity with Gaussian smoothing around the theoretical dispersion relation
    if switch:
        dkt = kt - kt_theory
        gauss_spectrum = ne.evaluate('exp(-dkt**2 / gauss_width ** 2)')
    else:
        gauss_spectrum = np.zeros(kt.shape)

    return gauss_spectrum



def spectrum_downsample(
        measured_spectrum: np.ndarray,
        res: float,
        fps: float,
        window_dims : Tuple[int, int, int],
        downsample: int,
) -> np.ndarray:
    """
    trims the measured spectrum reducing its dimensions [1] and [2] bny a factor "downsample"

    Parameters
    ----------
    measured_spectrum : np.ndarray
        measured, averaged, and normalised 3D power spectrum calculated with spectral.py

    res: float
        image resolution (m/pxl)

    fps: float
        image acquisition rate (fps)
    
    window_dims: [int, int, int]
        [dim_t, dim_y, dim_x] window dimensions
    
    downsample: int=1
        downsampling rate. If downsample > 1, then the spectrum is trimmed using a trimming ratio equal to 'downsample'.
        Trimming removes the high-wavenumber tails of the spectrum, which corresponds to downsampling the images spatially.

    Returns
    -------
    trimmed_spectrum : np.ndarray
        trimmed spectrum

    """
    
    kt_old, ky_old, kx_old = spectral.wave_numbers(window_dims,res,fps)
    
    res = res*downsample
    fps = fps
    
    window_dims = [window_dims[0], np.int64(np.ceil(window_dims[1]/downsample)), np.int64(np.ceil(window_dims[2]/downsample))]
    kt_new, ky_new, kx_new = spectral.wave_numbers(window_dims,res,fps)
    
    # this could probably be simplified, but it is to ensure that the trimmed wavenumber arrays are correctly 
    # aligned with the untrimmed one, since they will be calculated only based on res and fps in the following
    kx_indx_first = np.where(np.isclose(kx_old, np.min(kx_new), atol = 1e-06))[0][0]
    kx_indx_last = np.where(np.isclose(kx_old, np.max(kx_new), atol = 1e-06))[0][0]
    kx_indx = np.arange(kx_indx_first, kx_indx_last + 1)
    ky_indx_first = np.where(np.isclose(ky_old, np.min(ky_new), atol = 1e-05))[0][0]
    ky_indx_last = np.where(np.isclose(ky_old, np.max(ky_new), atol = 1e-05))[0][0]
    ky_indx = np.arange(ky_indx_first, ky_indx_last + 1)
    kt_indx_first = np.where(np.isclose(kt_old, np.min(kt_new), atol = 1e-05))[0][0]
    kt_indx_last = np.where(np.isclose(kt_old, np.max(kt_new), atol = 1e-05))[0][0]
    kt_indx = np.arange(kt_indx_first, kt_indx_last + 1)
    
    if len(kx_indx) != len(kx_new) | len(ky_indx) != len(ky_new) | len(kt_indx) != len(kt_new):
        raise ValueError("The dimensions of the trimmed array do not match the target after resampling")
        
    # trim the spectrum
    trimmed_spectrum = measured_spectrum[np.ix_(kt_indx,ky_indx,kx_indx)]
    
    
    return trimmed_spectrum, res, fps, window_dims



import numpy as np
from typing import Tuple, Literal


def sliding_window_idx(
    image: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (32, 32),
) -> np.ndarray:
    """Create y and x indices per interrogation window.

    Parameters
    ----------
    image : np.ndarray
        black-white image or template
    window_size : tuple
        size of interrogation window (y, x)
    overlap : tuple
        overlap of pixels of interrogation windows (y, x)

    Returns
    -------
    win_x : np.ndarray (w * y * x)
        x-indices of interrogation windows (w)
    win_y : np.ndarray (w * y * x)
        y-indices of interrogation windows (w)
    """

    x, y = get_rect_coordinates(
        image.shape,
        window_size,
        overlap,
        center_on_field=False
    )
    xi, yi = np.meshgrid(x, y)
    xi = (xi - window_size[1] // 2).astype(int)
    yi = (yi - window_size[0] // 2).astype(int)
    xi, yi = np.reshape(xi, (-1, 1, 1)), np.reshape(yi, (-1, 1, 1))

    win_x, win_y = np.meshgrid(
        np.arange(0, window_size[1]),
        np.arange(0, window_size[0])
    )
    win_x = win_x[np.newaxis, :, :] + xi
    win_y = win_y[np.newaxis, :, :] + yi
    return win_x, win_y


def sliding_window_array(
    image: np.ndarray,
    win_x: np.ndarray,
    win_y: np.ndarray,
) -> np.ndarray:

    return image[win_y, win_x]


def multi_sliding_window_array(
    imgs: np.ndarray,
    win_x: np.ndarray,
    win_y: np.ndarray,
    swap_time_dim=False,
) -> np.ndarray:
    """Get multiple interrogation windows from a stack of images lazily."""
    windows = np.stack(
        [
            sliding_window_array(
                img,
                win_x,
                win_y
            ) for img in imgs
        ]
    )

    if swap_time_dim:
        return np.swapaxes(windows, 0, 1)

    return windows


def get_axis_shape(
    dim_size: int,
    window_size: int,
    overlap: int,
) -> int:
    """
    get shape of image axis given its dimension size

    Parameters
    ----------
    dim_size : int
        size of axis [pix]
    window_size : int
        size of interrogation window over axis dimension [pix]
    overlap : int
        size of overlap [pix]

    Returns
    -------
    int, amount of interrogation windows over provided axis

    """
    axis_shape = (dim_size - window_size) // (window_size - overlap) + 1
    return axis_shape


def get_array_shape(
    dim_sizes: Tuple[int, int],
    window_sizes: Tuple[int, int],
    overlaps: Tuple[int, int]
):
    """
    Get the resulting shape of velocimetry results as a tuple of dimension sizes

    Parameters
    ----------
    dim_sizes : [int, int]
        sizes of axes [pix]
    window_sizes : [int, int]
        sizes of interrogation windows [pix]
    overlaps : [int, int]
        sizes of overlaps [pix]

    Returns
    -------
    shape of array returned from velocimetry analysis

    """
    array_shape = tuple(
        get_axis_shape(dim_size, window_size, overlap) for (
            dim_size,
            window_size,
            overlap
        ) in zip(
            dim_sizes,
            window_sizes,
            overlaps
        )
    )
    return array_shape


def get_axis_coords(
    dim_size: int,
    window_size: int,
    overlap: int,
    center_on_field: bool = False,
):
    """
    Get axis coordinates for one axis with provided dimensions and window size
    parameters. Overlap for windows can be provided.

    Parameters
    ----------
    dim_size : int
        size of axis [pix]
    window_size : int
        size of interrogation window over axis dimension [pix]
    overlap : int
        size of overlap [pix]
    center_on_field : bool, optional
        take the center of the window as coordinate (default, False)

    Returns
    -------
    x- or y-coordinates of resulting velocimetry grid

    """
    # get the amount of expected coordinates
    ax_shape = get_axis_shape(dim_size, window_size, overlap)
    coords = np.arange(ax_shape) * (window_size - overlap) + (window_size) / 2.0
    if center_on_field is True:
        coords_shape = get_axis_shape(
            dim_size=dim_size,
            window_size=window_size,
            overlap=overlap
        )
        coords += (dim_size - 1 - ((coords_shape - 1) * (window_size - overlap) + (
                window_size - 1))
        ) // 2
    return coords


def get_rect_coordinates(
    dim_sizes: Tuple[int, int],
    window_sizes: Tuple[int, int],
    overlap: Tuple[int, int],
    center_on_field: bool = False,
):
    """
    create meshgrid coordinates (x, y) of velocimetry results. Overlap can be
    provided in case each interrogation window is to overlap with the 
    neighbouring interrogation window.

    Parameters
    ----------
    dim_sizes : [int, int]
        sizes of axes [pix]
    window_sizes : [int, int]
        sizes of interrogation windows [pix]
    overlaps : [int, int]
        sizes of overlaps [pix]
    center_on_field : bool, optional
        take the center of the window as coordinate (default, False)

    Returns
    -------
    xi, yi: np.ndarray (2D), np.ndarray (2D)
        x- and y-coordinates in meshgrid form

    """
    y = get_axis_coords(
        dim_sizes[0],
        window_sizes[0],
        overlap[0],
        center_on_field=center_on_field
    )
    x = get_axis_coords(
        dim_sizes[1],
        window_sizes[1],
        overlap[1],
        center_on_field=center_on_field
    )

    return x, y

def normalize(
    imgs: np.ndarray,
    mode: Literal["xy", "time"] = "time"
):
    """
    normalizes images assuming the last two dimensions contain the x/y image
    intensities

    Parameters
    ----------
    imgs : np.ndarray (n x Y x X) or (n x m x Y x X)
        input images, organized in at least one stack
    Returns
    -------

    imgs_norm : np.ndarray (n x Y x X) or (n x m x Y x X)
        output normalized images, organized in at least one stack, 
        similar to imgs
    """
    # compute means and stds
    if mode == "xy":
        imgs_std = np.expand_dims(
            imgs.reshape(imgs.shape[0], imgs.shape[1], -1).std(axis=-1),
            axis=(-1, -2)
        )
        imgs_mean = np.expand_dims(
            imgs.reshape(imgs.shape[0], imgs.shape[1], -1).mean(axis=-1),
            axis=(-1, -2)
        )
    elif mode == "time":
        imgs_std = np.expand_dims(
            imgs.std(axis=-3),
            axis=-3
        )
        imgs_mean = np.expand_dims(
            imgs.mean(axis=-3),
            axis=-3
        )
    else:
        raise ValueError(f'mode must be "xy" or "time", but is "{mode}"')
    # this step takes a lot of memory, consider doing a stepwise replacement through numba instead of numpy in one go
    with np.errstate(divide='ignore', invalid='ignore'):
        imgs = (imgs - imgs_mean) / imgs_std
    imgs = np.nan_to_num(imgs)
    return imgs




"""File I/O functions for IWaVE."""

import glob
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import os

from PIL import Image
from tqdm import tqdm
from typing import Optional, Literal

try:
    import cv2
    ENABLE_CV2 = True
except:
    ENABLE_CV2 = False


def get_video(fn: str, start_frame: int = 0, end_frame: int = 4, stride: int = 1):
    """Read video frames from file

    Parameters
    ----------
    fn : str
        Path to the video file.
    start_frame : int, optional
        The starting frame number from which to begin extraction (default is 0).
    end_frame : int, optional
        The frame number at which to stop extraction (default is 4).
    stride : int, optional
        lower the sampling rate by this factor. Default 1.

    Returns
    -------
    np.ndarray
        3D array with images stacked over first dimension

    """
    try:
        import cv2
    except ImportError:
        raise ImportError("This function needs cv2. Install iwave with pip install iwave[extra]")
    if os.path.isfile(fn) is False:
        raise FileNotFoundError(
            f"The provided file path {fn} does not exist. Please check the path and try again."
        )
    cap = cv2.VideoCapture(fn)
    cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)

    # retrieve images from start to end frame
    return np.stack(
        [cv2.cvtColor(cap.read()[-1][::stride, ::stride], cv2.COLOR_BGR2GRAY) for _ in tqdm(range(end_frame - start_frame))]
    )
    

def get_imgs(path: str, wildcard: Optional[str] = None):
    """Read frames from images.

    Parameters
    ----------
    path : str
        Directory path where image files are stored.
    wildcard : Optional[str], default: None
        Optional wildcard pattern to match specific files within the directory.

    Returns
    -------
    np.ndarray
        3D array with images stacked over first dimension
    """
    # list all files in folder
    if wildcard:
        path = os.path.join(path, wildcard)

    fns = glob.glob(path)
    fns.sort()
    # read in memory as black-and-white
    return np.stack(
        [np.array(Image.open(fn).convert('L')) for fn in tqdm(fns)])


def plot_spectrum(
    spectrum: np.ndarray,
    kt: np.ndarray,
    ky: np.ndarray,
    kx: np.ndarray,
    dim: Literal["x", "y", "time"],
    slice: Optional[int] = None,
    log: bool = True,
    ax: Optional[matplotlib.axes.Axes] = None,
    **kwargs
):
    """Plot 2D slice of spectrum of a window.

    Parameters
    ----------
    spectrum : np.ndarray
        The spectrum data array.
    kt : np.ndarray
        Array of wave numbers or time indices corresponding to the first (time) dimension.
    ky : np.ndarray
        Array of wave numbers corresponding to the second (x) dimension.
    kx : np.ndarray
        Array of wave numbers corresponding to the third (y) dimension.
    dim : Literal["x", "y", "time"]
        Dimension along which to slice the spectrum for plotting ("x", "y", or "time").
    slice : int, optional
        Index along the specified dimension to slice the spectrum. If not provided the middle index is used.
    log : bool, optional
        If True (default), spectrum is plotted on log scale.
    ax : matplotlib.axes.Axes, optional
        Matplotlib Axes object to plot on. If None, a new Axes object is created.
    kwargs : dict
        Additional arguments passed to the plotting function `matplotlib.pyplot.pcolormesh`.

    Returns
    -------
    matplotlib.collections.QuadMesh
        matplotlib handle to mesh plot
    """
    if ax is None:
        ax = plt.axes()
    match dim:
        case "x":
            xlabel = "ky"
            ylabel = "kt"
            xvals = kt
            yvals = ky
            slice_vals = kx
            if slice is None:
                # define middle of dimension
                slice = spectrum.shape[-1] // 2
            imvals = spectrum[:, :, slice]
        case "y":
            xlabel = "kx"
            ylabel = "kt"
            xvals = kt
            yvals = kx
            slice_vals = ky
            if slice is None:
                # define middle of dimension
                slice = spectrum.shape[-2] // 2
            imvals = spectrum[:, slice, :]
        case "time":
            xlabel = "ky"
            ylabel = "kx"
            xvals = ky
            yvals = kx
            slice_vals = kt
            if slice is None:
                # define middle of dimension
                slice = spectrum.shape[-3] // 2
            imvals = spectrum[slice, :]
        case _:
            raise ValueError(f"Invalid dimension: {dim}")
    if log:
        imvals = np.log(imvals)
    p = ax.pcolormesh(yvals, xvals, imvals, **kwargs)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(f"Spectrum {dim} wn: {round(slice_vals[slice], 3)}")
    return p


def plot_spectrum_fitted(
    spectrum: np.ndarray,
    kt_waves_theory: np.ndarray, 
    kt_advected_theory: np.ndarray, 
    kt: np.ndarray,
    ky: np.ndarray,
    kx: np.ndarray,
    dim: Literal["x", "y", "time"],
    slice: Optional[int] = None,
    log: bool = True,
    ax: Optional[matplotlib.axes.Axes] = None,
    **kwargs
):
    """Plot 2D slice of spectrum of a window.

    Parameters
    ----------
    spectrum : np.ndarray
        The spectrum data array.
    kt_waves_theory : np.ndarray
        Theoretical frequency of gravity waves according to optimised velocity vector.
    kt_advected_theory : np.ndarray
        Theoretical frequency of turbulence-driven waves according to optimised velocity vector.
    kt : np.ndarray
        Array of wave numbers or time indices corresponding to the first (time) dimension.
    ky : np.ndarray
        Array of wave numbers corresponding to the second (x) dimension.
    kx : np.ndarray
        Array of wave numbers corresponding to the third (y) dimension.
    dim : Literal["x", "y", "time"]
        Dimension along which to slice the spectrum for plotting ("x", "y", or "time").
    slice : int, optional
        Index along the specified dimension to slice the spectrum. If not provided the middle index is used.
    log : bool, optional
        If True (default), spectrum is plotted on log scale.
    ax : matplotlib.axes.Axes, optional
        Matplotlib Axes object to plot on. If None, a new Axes object is created.
    kwargs : dict
        Additional arguments passed to the plotting function `matplotlib.pyplot.pcolormesh`.

    Returns
    -------
    matplotlib.collections.QuadMesh
        matplotlib handle to mesh plot
    """
    if ax is None:
        ax = plt.axes()
    match dim:
        case "x":
            xlabel = "ky"
            ylabel = "kt"
            xvals = kt
            yvals = ky
            slice_vals = kx
            if slice is None:
                # define middle of dimension
                slice = spectrum.shape[-1] // 2
            imvals = spectrum[:, :, slice]
            linevals_waves = np.squeeze(kt_waves_theory[0,:,slice])
            linevals_advected = np.squeeze(kt_advected_theory[0,:,slice])
        case "y":
            xlabel = "kx"
            ylabel = "kt"
            xvals = kt
            yvals = kx
            slice_vals = ky
            if slice is None:
                # define middle of dimension
                slice = spectrum.shape[-2] // 2
            imvals = spectrum[:, slice, :]
            linevals_waves = np.squeeze(kt_waves_theory[0,slice,:])
            linevals_advected = np.squeeze(kt_advected_theory[0,slice,:])
        case "time":
            xlabel = "ky"
            ylabel = "kx"
            xvals = ky
            yvals = kx
            slice_vals = kt
            if slice is None:
                # define middle of dimension
                slice = spectrum.shape[-3] // 2
            imvals = spectrum[slice, :]
        case _:
            raise ValueError(f"Invalid dimension: {dim}")
    if log:
        imvals = np.log(imvals)
    p = ax.pcolormesh(yvals, xvals, imvals, **kwargs)
    ax.plot(yvals,linevals_waves, color='red')
    ax.plot(yvals,linevals_advected, color='red')
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(f"Spectrum {dim} wn: {round(slice_vals[slice], 3)}")
    ax.set_ylim([0, np.max(xvals)])
    return p


import iwave
import sphinx_autosummary_accessors

# -- Project information -----------------------------------------------------

project = 'IWaVE'
copyright = '2024, Giulio Dolcetti, Salvador Pena-Haro, Hessel Winsemius'
author = 'Giulio Dolcetti, Salvador Pena-Haro, Hessel Winsemius'
release = iwave.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autosummary_accessors",
    "sphinx_design"
]
autosummary_generate = True
templates_path = ["_templates", sphinx_autosummary_accessors.templates_path]

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

source_suffix = ".rst"
master_doc = "index"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
autodoc_member_order = "bysource"
autoclass_content = "both"

html_static_path = ['_static']
# No specific theme applied yet. Consider altering the theme css below.
# html_css_files = ["some-theme-iwave.css"]

html_theme_options = {
    "show_nav_level": 2,
    "navbar_align": "content",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/DataForWater/IWaVE",  # required
            "icon": "https://upload.wikimedia.org/wikipedia/commons/9/91/Octicons-mark-github.svg",
            "type": "url",
        },
        {
            "name": "Data4Water",
            "url": "https://github.com/DataForWater",
            "icon": "_static/logo.jpg",
            "type": "local",
        },
    ],
}

html_context = {
    "github_url": "https://github.com",
    "github_user": "DataForWater",
    "github_repo": "IWaVE",
    "github_version": "docs",
    "doc_path": "docs",
}


remove_from_toctrees = ["_generated/*", "_build/doctrees/*"]




import numpy as np
from iwave import dispersion, spectral

def test_dispersion(img_windows_norm):
    """Check shape of dispersion result."""
    kt, ky, kx = spectral.wave_numbers(img_windows_norm.shape, res=0.01, fps=25)
    kt_gw, kt_turb = dispersion.dispersion(ky, kx, velocity=[1, 0], depth=1, vel_indx=1)
    # test if the size of kt_gw is 1 x dim_y x dim_x
    assert kt_gw.shape == (
        (1, img_windows_norm.shape[-2], img_windows_norm.shape[-1])
    )

def test_intensity(img_windows_norm):
    """Check shape of intensity result."""
    kt, ky, kx = spectral.wave_numbers(img_windows_norm.shape, res=0.01, fps=25)
    kt_gw, kt_turb = dispersion.dispersion(ky, kx, velocity=[1, 0], depth=1, vel_indx=1)
    th_spectrum = dispersion.theoretical_spectrum(kt_gw, kt_turb, kt, gauss_width=1, gravity_waves_switch=True, turbulence_switch=True)
    measured_spectrum = spectral._numpy_fourier_transform(img_windows_norm)
    # test if the size of the theoretical spectrum is equal to the size of the spectrum measured with numba
    assert th_spectrum.shape == (
    (int(np.ceil(measured_spectrum.shape[-3] / 2)), measured_spectrum.shape[-2], measured_spectrum.shape[-1])
    )
        
    
def test_downsample(img_windows_norm):
    """Check shape of intensity result."""
    img_single_window = img_windows_norm[0]
    full_spectrum = spectral._numpy_fourier_transform(img_single_window)
    trimmed_spectrum, res, fps, window_dims = dispersion.spectrum_downsample(full_spectrum, res=0.01, fps=25, window_dims=img_single_window.shape, downsample=2)
    downsampled_img = img_single_window[:, ::2, ::2]
    downsampled_spectrum = spectral._numpy_fourier_transform(downsampled_img)
    assert trimmed_spectrum.shape == downsampled_spectrum.shape


"""IWaVE: Image-based Wave Velocimetry Estimation"""

__version__ = "0.3.1"

import os

CONCURRENCY = os.environ.get("IWAVE_NUM_THREADS", None)
if CONCURRENCY is not None:
    CONCURRENCY = int(CONCURRENCY)

from . import const
from . import dispersion
from . import io
from . import sample_data
from . import spectral
from . import window
from .data_models import LazySpectrumArray, LazyWindowArray
from . import optimise
from .iwave import Iwave




"""Test fixtures."""
import numpy as np
import pytest

from iwave import sample_data, io


@pytest.fixture
def fn_video():
    return sample_data.get_sheaf_dataset()


@pytest.fixture
def imgs(fn_video):
    """ 4 selected frames from sample dataset, read with reader helper function.
    Result is [4 x n x m ] np.ndarray """
    return io.get_video(fn_video)


# @pytest.fixture
# def img_windows(fn_windows):
#     with open(fn_windows, "rb") as f:
#         windows = np.load(f)
#     return windows

@pytest.fixture
def img_windows():
    fn_windows = sample_data.get_sheaf_windows()
    with open(fn_windows, "rb") as f:
        windows = np.load(f)
    return windows


@pytest.fixture
def img_windows_norm(img_windows):
    img_windows = img_windows - img_windows.mean(axis=0)
    img_windows = img_windows / img_windows.std(axis=0)
    img_windows[np.isinf(img_windows)] = 0
    img_windows[np.isnan(img_windows)] = 0
    return img_windows




"""Retrieval of sample dataset."""
import os.path
import numpy as np
from iwave import window, io


def get_sheaf_dataset():
    """Retrieve and cache sample dataset of Sheaf river."""
    try:
        import pooch
    except:
        raise ImportError("This function needs pooch. Install iwave with pip install iwave[extra]")
    # Define the DOI link
    filename = "Fersina_20230630.avi"
    base_url = "doi:10.5281/zenodo.13935859"
    url = base_url + "/" + filename
    print(f"Retrieving or providing cached version of dataset from {url}")
    # Create a Pooch registry to manage downloads
    registry = pooch.create(
        # Define the cache directory
        path=pooch.os_cache("iwave"),
        # Define the base URL for fetching data
        base_url=base_url,
        # Define the registry with the file we're expecting to download
        registry={
            filename: None
        },
    )
    # Fetch the dataset
    file_path = registry.fetch(filename, progressbar=True)
    print(f"Sheaf dataset is available in {file_path}")
    return file_path


def get_sheaf_windows(start_frame=0, end_frame=500, window_size=(64, 64), overlap=(32, 32), max_windows=4):
    src_path = get_sheaf_dataset()
    dst_path = os.path.join(
        os.path.split(src_path)[0], "windows_{:04d}_{:04d}.bin".format(start_frame, end_frame)
    )
    if not os.path.exists(dst_path):
        imgs = io.get_video(src_path, start_frame=start_frame, end_frame=end_frame)
    # get the x and y coordinates per window
        win_x, win_y = window.sliding_window_idx(imgs[0], window_size=window_size, overlap=overlap)
        # apply the coordinates on all images
        window_stack = window.multi_sliding_window_array(
            imgs,
            win_x,
            win_y,
            swap_time_dim=True
        )
        # only select the ones we are interested in
        window_stack = window_stack[0:max_windows]

        with open(dst_path, "wb") as f:
            np.save(f, window_stack)
    return dst_path


import numpy as np
import pytest

from iwave import spectral


def test_get_wave_numbers(img_windows, res=0.02, fps=25):
    """Check shape of wave numbers, conditional on shape of image."""
    # feed in only the first image window, and of that, only the first frame
    kt, ky, kx = spectral.wave_numbers(img_windows[0].shape, res=res, fps=fps)
    assert len(kt) == int(np.ceil(len(img_windows[0])/2))
    assert len(ky) == img_windows.shape[-2]
    assert len(kx) == img_windows.shape[-1]


def test_numpy_fft(img_windows_norm):
    """Check shape of spectrum."""
    windows = img_windows_norm[-1]
    spectrum = spectral._numpy_fourier_transform(windows)
    assert spectrum.shape == (
        int(np.ceil(len(windows) / 2)),
        windows.shape[1],
        windows.shape[2]
    )


def test_numba_fft(img_windows_norm):
    """Compare spectrum derived with numba versus derived with numpy."""
    spectrum = spectral._numba_fourier_transform(img_windows_norm[-1])
    spectrum_numpy = spectral._numpy_fourier_transform(img_windows_norm[-1])
    # test if the answer is (very) close to the answer of the numpy version
    assert np.allclose(spectrum_numpy, spectrum)


def test_numba_fft_multi(img_windows_norm):
    """Compare multi-spectrum derived with numba versus derived with numpy."""
    spectra = spectral._numba_fourier_transform_multi(img_windows_norm)
    # test if all image windows give the same result as the numpy version of the spectrum code
    for windows, spectrum in zip(img_windows_norm, spectra):
        spectrum_numpy = spectral._numpy_fourier_transform(windows)
        assert np.allclose(spectrum_numpy, spectrum)


def test_sliding_window_spectrum(img_windows_norm):
    """Check shape of average of multiple spectra over several time slices."""
    spectrum = spectral.sliding_window_spectrum(img_windows_norm, 20, 10)
    # test if the spectra have the desired size
    assert spectrum.shape[-3] == int(np.ceil(20 / 2))


"""IWaVE main api."""

import cv2
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

from typing import Optional, Tuple, Literal
from iwave import window, spectral, io, optimise, dispersion, LazySpectrumArray, LazyWindowArray


repr_template = """
Resolution [m]: {}
Window size (y, x): {}
Overlap (y, x): {}
Size of time slices: {}
Overlap in time slices: {}
Number of images: {}
Frames per second: {}
""".format

# optim kwargs for differential evolution algorithm
OPTIM_KWARGS_SADE = {
    "strategy" : 'best1bin', 
    "popsize": 8,
    "maxiter": int(1e05),
    "workers": 1,
    "init" : 'sobol',
    "atol" : 1e-12
}

class Iwave(object):
    def __init__(
        self,
        resolution: float,
        window_size: Tuple[int, int] = (64, 64),
        overlap: Tuple[int, int] = (0, 0),
        time_size: int = 128,
        time_overlap: int = 0,
        fps: Optional[float] = None,
        imgs: Optional[np.ndarray] = None,
        norm: Optional[Literal["time", "xy"]] = "time",
        spectrum_threshold: Optional[float] = 1.0,
        smax: Optional[float] = 4.0,
        dmin: Optional[float] = 0.01,
        dmax: Optional[float] = 3.0,
        penalty_weight: Optional[float]=1,
        gravity_waves_switch: Optional[bool]=True,
        turbulence_switch: Optional[bool]=True,
        window_chunk_size: Optional[int] = 50,
    ):
        """Initialize an Iwave instance.

        Parameters
        ----------
        resolution : float
            Physical resolution of the images that will be provided.
        window_size : Tuple[int, int], optional
            Size (pixels y, pixels x) of interrogation windows over which velocities are estimated.
        overlap : Tuple[int, int], optional
            Overlap in space (y, x) used to select windows from images or frames.
        time_size : int, optional
            Amount of frames in time used for one spectral analysis. Must be <= amount of frames available.
        time_overlap : int, optional
            Amount of overlap in frames, used to establish time slices.
        fps : float, optional
            Frames per second, can be set at the start, otherwise inherited from read video, or imposed with image set.
        imgs : Optional[np.ndarray], optional
            Array of images used for analysis. If not provided, defaults to None.
        norm : Literal["time", "xy"]
            Normalization to apply over subwindowed images, either over time ("time") or space ("xy").
        spectrum_threshold : float, optional
            cut-off threshold for spectrum intensities, anything below this value is set to zero. Defaults to 1.0.
        smax : float, optional
            Maximum velocity expected in the scene. Defaults to 4 m/s
        dmin : float, optional
            Minimum depth expected in the scene. Defaults to 0.01 m
        dmax : float, optional
            Maximum depth expected in the scene. Defaults to 3 m
        penalty_weight : float, optional
            Parameter to reduce the risk of outliers by penalising solutions with high velocity modulus.
            Inactive if set to 0. Defaults to 1. 
            Outliers can be frequent if smax > 2 * flow velocity. Increase penalty_weight only if reducing smax 
            is not possible, since setting penalty_weight > 0 may introduce a bias.
        gravity_waves_switch: bool, optional
            If True, gravity waves are modelled. If False, gravity waves are NOT modelled. Default True. 
            Setting gravity_waves_swtich = False may improve performance if floating tracers dominate the scene and waves are minimal.
        turbulence_switch: bool=True
            If True, turbulence-generated patterns and/or floating particles are modelled. If False, 
            turbulence-generated patterns and/or floating particles are NOT modelled. Default True.
            Setting turbulence_switch = False may improve performance if water waves dominate the scene, or if tracers
            dynamics are not representative of the actual flow velocity (e.g., due to air resistance, surface tension, etc.)
        window_chunk_size : int, optional
            Number of windows to process at a time. Defaults to 50.
        """
        self.window_chunk_size = window_chunk_size
        self.resolution = resolution
        # ensures that window dimensions are even. this is to facilitate dimension reduction of the spectra.
        # this is currently working only for a downsampling rate of 2
        # TODO: generalise to any downsampling value
        self.window_size = tuple((dim if dim % 2 == 0 else dim + 1) for dim in window_size) 
        self.overlap = overlap
        self.time_size = time_size
        self.time_overlap = time_overlap
        self.norm = norm
        self.spectrum_threshold = spectrum_threshold
        self.smax = smax
        self.dmin = dmin
        self.dmax = dmax
        self.fps = fps
        self.penalty_weight = penalty_weight
        self.gravity_waves_switch = gravity_waves_switch
        self.turbulence_switch = turbulence_switch
        if imgs is not None:
            self.imgs = imgs
        else:
            self.imgs = None
        self.win_x = None
        self.win_y = None
        self.vy = None  # y velocity component (m/s)
        self.vx = None  # x velocity component (m/s)
        self.d = None  # water depth (m)
        self.cost = None  # cost function value (float)
        self.quality = None  # quality parameter (0 < q < 1), where 1 is highest quality and 0 is lowest quality
        self.status = None  # Boolean flag indicating if the optimizer exited successfully
        self.message = None  # termination message returned by the optimiser
                
    def __repr__(self):
        if self.imgs is not None:
            no_imgs = len(self.imgs)
        else:
            no_imgs = None

        return repr_template(
            self.resolution,
            self.window_size,
            self.overlap,
            self.time_size,
            self.time_overlap,
            no_imgs,
            self.fps
        )

    @property
    def imgs(self):
        """Return image set."""
        return self._imgs

    @imgs.setter
    def imgs(self, images):
        """Set images and derived properties subwindows, x and y axes, wave number axes and derived spectra."""
        if images is not None:
            if images.ndim != 3:
                raise ValueError(f"Provided image array must have 3 dimensions. Provided dimensions are {images.ndim}: {images.shape}")
        self._imgs = images
        if images is not None:
            # TODO: check if image set is large enough for the given dimension of subwindowing and time windowing
            # subwindow images and get axes. This always necessary, so in-scope methods only.
            # self._get_subwindow(images)
            # self._get_subwindow(da.from_array(images, chunks=(len(images), -1, -1)))
            self._get_win_xy(images[0])  # set sampling indexes per interrogation window
            self._get_x_y_axes(images)  # set envisaged axes
            # set the wave numbers from image fps
            self._get_wave_numbers()
            self._windows = LazyWindowArray(
                self.imgs,
                sliding_window_func=window.multi_sliding_window_array,
                win_x=self.win_x,
                win_y=self.win_y,
                norm=self.norm
            )
            self._spectrum = LazySpectrumArray(
                windows=self.windows,
                time_size=self.time_size,
                time_overlap=self.time_overlap,
                kt=self.kt,
                kx=self.kx,
                ky=self.ky,
                smax=self.smax,
                threshold=self.spectrum_threshold
            )

    @property
    def spectrum(self):
        """Return images represented in subwindows."""
        if not hasattr(self, "_spectrum"):
            if self.imgs is None:
                return None
            else:
                self._spectrum = LazySpectrumArray(
                    windows=self.windows,
                    time_size=self.time_size,
                    time_overlap=self.time_overlap,
                    kt=self.kt,
                    kx=self.kx,
                    ky=self.ky,
                    smax=self.smax,
                    threshold=self.spectrum_threshold
                )
        return self._spectrum

    @spectrum.setter
    def spectrum(self, _spectrum):
        """Set images represented in subwindows directly, only used for testing purposes."""
        self._spectrum = _spectrum

    @property
    def windows(self):
        """Return images represented in subwindows."""
        if not hasattr(self, "_windows"):
            if self.imgs is None:
                return None
            else:
                self._windows = LazyWindowArray(
                    self.imgs,
                    sliding_window_func=window.multi_sliding_window_array,
                    win_x=self.win_x,
                    win_y=self.win_y,
                    norm=self.norm
                )
        return self._windows

    @property
    def x(self):
        """Return x-axis of velocimetry field."""
        return self._x

    @x.setter
    def x(self, _x):
        self._x = _x

    @property
    def y(self):
        """Return y-axis of velocimetry field."""
        return self._y

    @y.setter
    def y(self, _y):
        self._y = _y

    @property
    def win_x(self):
        return self._win_x

    @win_x.setter
    def win_x(self, win_x):
        self._win_x = win_x

    @property
    def win_y(self):
        return self._win_y

    @win_y.setter
    def win_y(self, win_y):
        self._win_y = win_y

    @property
    def spectrum_dims(self):
        """Return expected dimensions of the spectrum derived from image windows."""
        return (self.time_size, *self.window_size)

    def _get_win_xy(self, img_sample):
        win_x, win_y = window.sliding_window_idx(
            img_sample,
            window_size=self.window_size,
            overlap=self.overlap,
        )
        self.win_x = win_x
        self.win_y = win_y

    def _get_wave_numbers(self):
        """Prepare and set wave number axes."""
        self.kt, self.ky, self.kx = spectral.wave_numbers(
            self.spectrum_dims,
            self.resolution, self.fps
        )

    def _get_x_y_axes(self, images: np.ndarray):
        """Prepare and set x and y axes of velocity grid."""
        x, y = window.get_rect_coordinates(
            dim_sizes=images.shape[-2:],
            window_sizes=self.window_size,
            overlap=self.overlap,
        )
        self.x = x
        self.y = y

    def plot_spectrum(
        self,
        window_idx: int,
        dim: Literal["x", "y", "time"],
        slice: Optional[int] = None,
        log: bool = True,
        ax: Optional[matplotlib.axes.Axes] = None,
        **kwargs
    ):
        """Plot 2D slice of spectrum of selected subwindow.

        Parameters
        ----------
        window_idx : int
            Index of the spectrum window to plot.
        dim : {"x", "y", "time"}
            Dimension along which to plot the spectrum.
        slice : int, optional
            Index of the slice to plot in the specified dimension. If not provided, the middle index is used.
        log : bool, optional
            If True (default), spectrum is plotted on log scale.
        ax : matplotlib.axes.Axes, optional
            Matplotlib Axes object to plot on. New axes will be generated if not provided.
        kwargs
            Additional keyword arguments to pass to the plotting function pcolormesh.
            See :py:func:`matplotlib.pyplot.pcolormesh` for options.
        """
        spectrum_sel = self.spectrum[window_idx]
        p = io.plot_spectrum(spectrum_sel, self.kt, self.ky, self.kx, dim, slice, ax=ax, log=log, **kwargs)
        return p
    
    def plot_spectrum_fitted(
        self,
        window_idx: int,
        dim: Literal["x", "y", "time"],
        slice: Optional[int] = None,
        log: bool = True,
        ax: Optional[matplotlib.axes.Axes] = None,
        **kwargs
        ):
        """Plot 2D slice of spectrum of selected subwindow.

        Parameters
        ----------
        window_idx : int
            Index of the spectrum window to plot.
        dim: {"x", "y", "time"}
            Dimension along which to plot the spectrum.
        slice : int, optional
            Index of the slice to plot in the specified dimension. If not provided, the middle index is used.
        log : bool, optional
            If True (default), spectrum is plotted on log scale.
        ax : matplotlib.axes.Axes, optional
            Matplotlib Axes object to plot on. New axes will be generated if not provided.
        kwargs
            Additional keyword arguments to pass to the plotting function pcolormesh.
            See :py:func:`matplotlib.pyplot.pcolormesh` for options.
        """
        spectrum_sel = self.spectrum[window_idx]
        kt_waves_theory, kt_advected_theory = dispersion.dispersion(
            self.ky,
            self.kx,
            (self.vy.flatten()[window_idx], self.vx.flatten()[window_idx]),
            depth=1,
            vel_indx=0.85
        )
        p = io.plot_spectrum_fitted(
            spectrum_sel,
            kt_waves_theory,
            kt_advected_theory,
            self.kt,
            self.ky,
            self.kx,
            dim,
            slice,
            ax=ax,
            log=log,
            **kwargs
        )
        return p

    def read_imgs(self, path: str, fps: float, wildcard: str = None):
        """Read frames stored as images on disk from path and wildcard.

        Parameters
        ----------
        path : str
            The directory path where the image frames are located.
        fps : float
           frames per second (must be explicitly set if it cannot be read from the video)
        wildcard : str, optional
            The pattern to match filenames. Defaults to None, meaning all files in the directory will be read.
        """
        self.fps = fps
        self.imgs = io.get_imgs(path=path, wildcard=wildcard)

    def read_video(self, file: str, start_frame: int = 0, end_frame: int = 4, stride: int = 1):
        """Read video from start to end frame.

        Parameters
        ----------
        file : str
            Path to the video file.
        start_frame : int, optional
            The starting frame number from which to begin reading the video.
        end_frame : int, optional
            The ending frame number until which to read the video.
        stride : int, optional
            lower the sampling rate by this factor. Default 1.
        """
        # set the FPS from the video metadata
        cap = cv2.VideoCapture(file)
        self.fps = cap.get(cv2.CAP_PROP_FPS)

        cap.release()
        del cap
        # Retrieve images
        self.imgs = io.get_video(fn=file, start_frame=start_frame, end_frame=end_frame, stride=stride)



    def save_frames(self, dst: str):
        raise NotImplementedError

    def save_windows(self):
        raise NotImplementedError


    def velocimetry(
        self,
        alpha: float = 0.85,
        depth: float = 1.,  # If depth = 0, then the water depth is estimated.
        twosteps: bool = False,
        **opt_kwargs # If True, the calculations are initially performed on a spectrum with reduced dimensions,
                            # and subsequently refined during a second step using the whole spectrum. This will reduce 
                            # computational time for large problems, but may reduce accuracy.
    ):
        """
        Estimate and set the velocity components u and v on the instance from the subwindowed spectra.

        The optimisation is performed using the differential evolution algorithm of `scipy.optimize`. You can pass
        arguments of this function to the optimiser. Default arguments are set as a starting point.

        If you set `twosteps=True`, the optimisation is performed twice, with a reduced spectrum in the first step
        without optimizing depth, and a refined step with full spectrum and optimizing depth.

        Parameters
        ----------
        alpha : float, optional
            depth-average to surface velocity ratio [-], default 0.85
        depth : float, optional
            depth of the water column [m], default 1. If set to 0. it will be estimated.
        twosteps : bool, optional
            if set, perform the optimisation twice, with a reduced spectrum in the first step without optimizing depth
            and full spectrum in the second step with optimizing depth. Default False.
        See also
        --------
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html

        """
        if self.spectrum is None:
            raise AttributeError(
                "No images available. Please set images first `iw.imgs = imgs`, where `imgs` is a numpy array "
                "of images, or read a video file using `iw.read_video(file)`")
        # ensure defaults are set if nothing is provided
        if not opt_kwargs:
            opt_kwargs = OPTIM_KWARGS_SADE
        # set search bounds to -/+ maximum velocity for both directions
        if depth == 0:  # If depth = 0, then the water depth is estimated.
            bounds = [(-self.smax, self.smax), (-self.smax, self.smax), (self.dmin, self.dmax)]
            if twosteps == False:
                self.penalty_weight = 0
                print(f"Depth estimation with the 1 step approach is inaccurate when penalty_weight is not zero. Now setting penalty_weight = 0. Consider reducing smax if results are incorrect, or use the two-steps approach.")
        else:
            bounds = [(-self.smax, self.smax), (-self.smax, self.smax), (depth, depth)]
        # Create a list of bounds for each window. This is to enable narrowing the bounds locally during multiple passages.
        bounds_list = [bounds for _ in range(len(self.spectrum))]
        
        if twosteps == True:
            bounds_firststep = bounds_list
            if depth==0: # for the first step, neglect water depth effects by assuming a large depth
                for i in range(len(bounds_list)):
                    bounds_firststep[i] = [bounds[0], bounds[1], (10, 10)]
            output_step1, _, _ = optimise.optimise_velocity(
                self.spectrum,
                bounds_firststep,
                alpha,
                self.spectrum_dims,
                self.resolution,
                self.fps,
                self.penalty_weight,  
                self.gravity_waves_switch, 
                self.turbulence_switch, 
                downsample=2, # for the first step, reduce the data size by 2
                gauss_width=1,  # TODO: figure out defaults
                desc="Optimizing windows 1st pass",
                **opt_kwargs
            )
            # re-initialise the problem using narrower bounds between 90% and 110% of the first step solution
            vy_step1 = output_step1[:, 0]
            vx_step1 = output_step1[:, 1]
            for i in range(len(bounds_list)):
                bounds_list[i] = [(vy_step1[i]-0.1*np.abs(vy_step1[i]), vy_step1[i]+0.1*np.abs(vy_step1[i])), 
                    (vx_step1[i]-0.1*np.abs(vx_step1[i]), vx_step1[i]+0.1*np.abs(vx_step1[i])), 
                        (bounds[2][0], bounds[2][1])]
            opt_kwargs["popsize"] = max(1, opt_kwargs["popsize"] // 2) # reduce the population size for the second step
            self.penalty_weight = 0 # set penalty_weight = 0 for the second step
        output, cost, quality = optimise.optimise_velocity(
            self.spectrum,
            bounds_list,
            alpha,
            self.spectrum_dims,
            self.resolution,
            self.fps,
            self.penalty_weight,  
            self.gravity_waves_switch, 
            self.turbulence_switch, 
            downsample=1,
            gauss_width=1,  # TODO: figure out defaults
            desc="Optimizing windows 2nd pass" if twosteps else "Optimizing windows",
            **opt_kwargs
        )
        self.vy = output[:, 0].reshape(len(self.y), len(self.x))
        self.vx = output[:, 1].reshape(len(self.y), len(self.x))
        self.d = output[:, 2].reshape(len(self.y), len(self.x))
        self.cost = cost.reshape(len(self.y), len(self.x))
        self.quality = quality.reshape(len(self.y), len(self.x))
        self.status = True
        self.message = "Optimization terminated successfully."
        
    
    def plot_velocimetry(self, ax: Optional[matplotlib.axes.Axes] = None, **kwargs):
        """Plot the estimated velocity components u and v on the axes instance."""
        if ax is None:
            ax = plt.axes()
        p = ax.quiver(self.x, self.y, self.vx, self.vy, **kwargs)
        return p



""" tests for window manipulations """
import numpy as np
import pytest

from iwave import window


def test_get_axis_shape(imgs):
    # get the last dimension (x-axis), assert if it fits
    dim_size = imgs.shape[-1]

    x_shape = window.get_axis_shape(
        dim_size=dim_size,
        window_size=64,
        overlap=32,
    )
    assert(x_shape == 10)


def test_get_array_shape(imgs):
    # get last two dimensions, assert numbers in returned dims
    dim_sizes = imgs.shape[-2:]
    xy_shape = window.get_array_shape(
        dim_sizes=dim_sizes,
        window_sizes=(64, 64),
        overlaps=(32, 32)
    )
    assert(xy_shape == (23, 10))


def test_get_axis_coords(imgs):
    dim_size = imgs.shape[-1]
    coords = window.get_axis_coords(
        dim_size,
        64,
        32,
    )
    assert(len(coords)==10)
    assert(np.allclose(np.array(coords[0:4]), np.array([32., 64., 96., 128.])))


def test_get_rect_coordinates(imgs):
    x, y = window.get_rect_coordinates(
        dim_sizes=imgs.shape[-2:],
        window_sizes=(64, 64),
        overlap=(32, 32),
    )
    # test first block of coords
    assert len(x) == 10
    assert len(y) == 23
    assert np.allclose(x[0:2], np.array([32., 64.]))
    assert np.allclose(y[0:2], np.array([32., 64.]))


def test_sliding_window_array(imgs):
    win_x, win_y = window.sliding_window_idx(imgs[0])
    img_wins = window.sliding_window_array(
        imgs[0],
        win_x,
        win_y
    )
    assert img_wins.shape == (23*10, 64, 64)


@pytest.mark.parametrize(
    ("swap_time_dim", "test_dims"),
    [
        (False, (4, 23*10, 64, 64)),
        (True, (23*10, 4, 64, 64))
    ]
)
def test_multi_sliding_window_array(imgs, swap_time_dim, test_dims):
    # get the x and y coordinates per window
    win_x, win_y = window.sliding_window_idx(imgs[0])
    # apply the coordinates on all images
    window_stack = window.multi_sliding_window_array(
        imgs,
        win_x,
        win_y,
        swap_time_dim=swap_time_dim
    )
    assert(window_stack.shape == test_dims)


def test_normalize(img_windows):
    img_norm = window.normalize(img_windows, mode="xy")
    # check if shape remains the same
    assert img_windows.shape == img_norm.shape
    # check if any window has mean / std of 0. / 1.
    assert np.isclose(img_norm[0][0].std(), 1.)
    assert np.isclose(img_norm[0][0].mean(), 0.)
    # check time normalization also
    img_norm = window.normalize(img_windows, mode="time")
    # check if random single time slice has mean / std of 0. / 1.
    assert np.isclose(img_norm[1, :, 1, 1].std(), 1.)
    assert np.isclose(img_norm[1, :, 1, 1].mean(), 0.)



import numpy as np
import pytest
import time
from iwave import spectral, dispersion, optimise, Iwave


def test_preprocessing():
    """Check shape of preprocessed spectrum against expected shape."""
    img = np.random.rand(3, 32, 32, 32)
    kt, ky, kx = spectral.wave_numbers(img.shape, res=0.02, fps=25)
    kt_gw, kt_turb = dispersion.dispersion(ky, kx, velocity=[1, 0], depth=1, vel_indx=1)
    synthetic_spectrum = dispersion.theoretical_spectrum(kt_gw, kt_turb, kt, gauss_width=1, gravity_waves_switch=True, 
                                                         turbulence_switch=True)
    measured_spectrum = spectral.sliding_window_spectrum(img, img.shape[1], 0)
    preprocessed_spectrum = spectral.spectrum_preprocessing(measured_spectrum, kt, ky, kx, velocity_threshold=5)
    #test if the size of the preprocessed spectrum matches the one of the theoretical spectrum
    assert preprocessed_spectrum[0].shape == synthetic_spectrum.shape

    
def test_nsp(img_size=(256, 64, 64), res=0.02, fps=25):
    """Compute nsp cost function, compare against known value."""
    kt, ky, kx = spectral.wave_numbers(img_size, res, fps)
    kt_gw, kt_turb = dispersion.dispersion(ky, kx, velocity=[1, 0], depth=1, vel_indx=1)
    synthetic_spectrum = dispersion.theoretical_spectrum(kt_gw, kt_turb, kt, gauss_width=1, gravity_waves_switch=True, 
                                                         turbulence_switch=True)
    cost = optimise.nsp_inv(synthetic_spectrum,synthetic_spectrum)
    expected_cost = np.sum(synthetic_spectrum)**2 / np.sum(synthetic_spectrum**2)
    #test if the auto-correlation matches the theoretical expectation based on a synthetic spectrum
    assert np.allclose(cost, expected_cost)

    
def test_cost_function_velocity_depth(img_size=(256, 64, 64), res=0.02, fps=25):
    """Check cost function (with depth) against expected gradients."""
    kt, ky, kx = spectral.wave_numbers(img_size, res, fps)
    depth_1 = 0.30
    depth_2 = 0.29
    depth_3 = 0.31
    velocity_y = 1
    velocity_x = 0
    params_1 = [velocity_y, velocity_x, np.log(depth_1)]
    params_2 = [velocity_y, velocity_x, np.log(depth_2)]
    params_3 = [velocity_y, velocity_x, np.log(depth_3)]
    vel_indx = 1
    kt_gw_1, kt_turb_1 = dispersion.dispersion(ky, kx, [velocity_y, velocity_x], depth_1, vel_indx)
    synthetic_spectrum_1 = dispersion.theoretical_spectrum(
        kt_gw_1,
        kt_turb_1,
        kt,
        gauss_width=1,
        gravity_waves_switch=True,
        turbulence_switch=True
    )
    cost_11 = optimise.cost_function_velocity_depth(
        params_1, synthetic_spectrum_1, vel_indx,
        img_size, res, fps, gauss_width=1, penalty_weight = 1,
        gravity_waves_switch=True, turbulence_switch=True
    )
    cost_12 = optimise.cost_function_velocity_depth(
        params_2, synthetic_spectrum_1, vel_indx,
        img_size, res, fps, gauss_width=1, penalty_weight = 1,
        gravity_waves_switch=True, turbulence_switch=True
    )
    cost_13 = optimise.cost_function_velocity_depth(
        params_3, synthetic_spectrum_1, vel_indx,
        img_size, res, fps, gauss_width=1, penalty_weight = 1,
        gravity_waves_switch=True, turbulence_switch=True
    )
    #test if the cost function increases when the depth deviates from optimal
    assert cost_12 > cost_11
    assert cost_13 > cost_11

# @pytest.mark.skip(reason="Optimization with depth is not yet stable")
def test_optimise_velocity_depth(img_size=(128, 64, 64), res=0.02, fps=12):
    """Check hypothetical case optimization with depth for one single window."""
    kt, ky, kx = spectral.wave_numbers(img_size, res, fps)
    velocity = [1, 0]
    depth = 0.2
    velocity_indx = 1
    kt_gw, kt_turb = dispersion.dispersion(
        ky,
        kx,
        velocity,
        depth,
        velocity_indx
    )
    synthetic_spectrum = dispersion.theoretical_spectrum(
        kt_gw,
        kt_turb,
        kt,
        gauss_width=1,
        gravity_waves_switch=True,
        turbulence_switch=True
    )
    # synthetic_spectrum = optimise.spectrum_preprocessing(synthetic_spectrum, kt, ky, kx, velocity_threshold=10, spectrum_threshold=1)
    # define ranges for optimization

    synthetic_spectrum = np.tile(synthetic_spectrum, (2,1,1,1)) # simulate multiple windows
    vel_y_min = 0
    vel_y_max = 2
    vel_x_min = -0.5
    vel_x_max = 0.5
    depth_min = 0.01
    depth_max = 1
    bounds = [[(vel_y_min, vel_y_max), (vel_x_min, vel_x_max), (depth_min, depth_max)]] * 2
    t1 = time.time()
    output, _, _ = optimise.optimise_velocity(
        synthetic_spectrum,
        bounds,
        velocity_indx,
        img_size,
        res,
        fps,
        gauss_width=1,
        penalty_weight=0,
        gravity_waves_switch=True,
        turbulence_switch=True,
        downsample=1,
        popsize=10,
        workers=1,
        maxiter=1000,
        # updating="deferred"
    )
    vel_y_optimal = output[:, 0]
    vel_x_optimal = output[:, 1]  
    depth_optimal = output[:, 2] 
    # print(f"Original velocity/depth was {velocity, depth}, optimized {optimal}")
    t2 = time.time()
    print(f"Took {t2 - t1} seconds")
    assert vel_x_max >= vel_x_min
    assert vel_y_max >= vel_y_min
    assert depth_max >= depth_min
    assert np.all(np.abs(vel_y_optimal - velocity[0]) < 0.01)
    assert np.all(np.abs(vel_x_optimal - velocity[1]) < 0.01)
    assert np.all(np.abs(depth_optimal - depth) < 0.05)
    
    
def test_iwave(img_size=(128, 64, 64), res=0.02, fps=12):
    """Check hypothetical case optimization with depth for one single window, using single and double pass"""
    np.random.seed(0)
    kt, ky, kx = spectral.wave_numbers(img_size, res, fps)
    velocity = [1, 0]
    depth = 0.2
    velocity_indx = 1
    depth_min = 0.01
    depth_max = 1
    
    iw = Iwave(
        resolution=res,
        window_size=(img_size[1], img_size[2]),
        overlap=(0, 0),
        time_size=128,
        time_overlap=0,
        fps=fps,
        dmin=depth_min,
        dmax=depth_max,
        gravity_waves_switch=True,
        turbulence_switch=True,
    )
    
    kt_gw, kt_turb = dispersion.dispersion(
        ky,
        kx,
        velocity,
        depth,
        velocity_indx
    )
    synthetic_spectrum = dispersion.theoretical_spectrum(
        kt_gw,
        kt_turb,
        kt,
        gauss_width=1,
        gravity_waves_switch=True,
        turbulence_switch=True
    )
    iw.spectrum = np.tile(synthetic_spectrum, (2,1,1,1)) # simulate multiple windows

    iw.x = np.array([0., 1.])
    iw.y = np.array([0.])
    iw.velocimetry(
        alpha=0.85,  # alpha represents the depth-averaged velocity over surface velocity [-]
        depth=depth,
        twosteps=False
    )
    
    vy_1step = iw.vy
    vx_1step = iw.vx 
    d_1step = iw.d
       
    iw.velocimetry(
        alpha=0.85,  # alpha represents the depth-averaged velocity over surface velocity [-]
        depth=depth,
        twosteps=True
    )

    vy_2steps = iw.vy
    vx_2steps = iw.vx 
    d_2steps = iw.d
    print("vy_1step:", vy_1step, "expected:", velocity[0])
    print("vx_1step:", vx_1step, "expected:", velocity[1])
    print("d_1step:", d_1step, "expected:", depth)
    print("vy_2steps:", vy_2steps, "expected:", velocity[0])
    print("vx_2steps:", vx_2steps, "expected:", velocity[1])
    print("d_2steps:", d_2steps, "expected:", depth)

    assert np.all(np.abs(vy_1step - velocity[0]) < 0.01)
    assert np.all(np.abs(vx_1step - velocity[1]) < 0.01)
    assert np.all(np.abs(d_1step - depth) < 0.05)
    
    assert np.all(np.abs(vy_2steps - velocity[0]) < 0.01)
    assert np.all(np.abs(vx_2steps - velocity[1]) < 0.01)
    assert np.all(np.abs(d_2steps - depth) < 0.05)
    


"""Constants."""

g = 9.81  # gravity acceleration (m/s^2)
surf_tens = 72.75e-3  # surface tension (N/m), ref. 20deg, from Vargaftik et al., 1983 - International tables of the
# surface tension of water, J.Phys.Chem.Ref.Data
density = 998.2  # water density (kg/m^3), ref. 20deg



"""Lazy data models for memory intensive processing."""
import numpy as np

from typing import Optional, Union

from iwave import window, spectral


class LazyWindowArray:
    def __init__(self, imgs, sliding_window_func, win_x, win_y, norm):
        """Initialize a lazy array-like object for subwindows.

        Parameters
        ----------
        imgs : np.ndarray
            Images (3D numpy array)
        sliding_window_func : callable
            Function that extracts subwindows
        win_x : np.ndarray
            Sampling indexes for window extraction in x direction
        win_y : np.ndarray
            Sampling indexes for window extraction in y direction
        norm : {'xy', 'time', None}
            Whether to normalize the windows and in what way
        """
        self.imgs = imgs
        self.sliding_window_func = sliding_window_func
        self.win_x = win_x
        self.win_y = win_y
        self.norm = norm  # Optional: normalization parameter

    def __repr__(self):
        return f"Interrogation windows dimensions (windows, time, y, x): {self.shape}"

    def __getitem__(self, idx):
        """
        On-the-fly generation of subwindows for the requested index or slice.
        """
        if isinstance(idx, slice):
            # Generate windows for a range of indices
            start, stop, step = idx.indices(len(self.win_x))
            win_x_sel = self.win_x[start:stop:step]
            win_y_sel = self.win_y[start:stop:step]
            # indexed_imgs = self.imgs[start:stop:step]
        elif isinstance(idx, Union[int, np.integer]):
            # Generate windows for a single index
            win_x_sel = self.win_x[idx : idx + 1]
            win_y_sel = self.win_y[idx : idx + 1]
        else:
            raise TypeError(f"Invalid index type: {type(idx)}. Only int or slice is supported.")

        # Use the sliding_window_func to generate the windows
        windows = self.sliding_window_func(self.imgs, win_x_sel, win_y_sel, swap_time_dim=True).astype(np.float64)

        # Apply normalization if needed
        if self.norm is not None:
            # find relevant windows with mean != 0. Only those require normalization
            nonzero_idx = np.where(np.mean(windows, axis=(1, 2, 3)) != 0)[0]
            if self.norm == "xy":
                windows[nonzero_idx] = window.normalize(windows[nonzero_idx], mode="xy")
            elif self.norm == "time":
                windows[nonzero_idx] = window.normalize(windows[nonzero_idx], mode="time")
        if isinstance(idx, slice):
            return windows
        elif isinstance(idx, int):
            return windows[0]
        else:
            raise TypeError(f"Invalid index type: {type(idx)}. Only int or slice is supported.")

    def __len__(self):
        """
        Return the total number of windows (or equivalent slices).
        """
        return len(self.win_x)

    @property
    def shape(self):
        return len(self), self.imgs.shape[0], self.win_x.shape[1], self.win_x.shape[2]


class LazySpectrumArray(object):
    def __init__(
        self,
        windows: LazyWindowArray,
        time_size: int,
        time_overlap: int,
        kt: np.ndarray,
        kx: np.ndarray,
        ky: np.ndarray,
        smax: Optional[float] = 4.0,
        threshold: Optional[float] = 1.0
    ):
        self.windows = windows
        self.time_size = time_size
        self.time_overlap = time_overlap
        self.kt = kt
        self.kx = kx
        self.ky = ky
        self.smax = smax
        self.threshold = threshold

    def __repr__(self):
        """Return string representation showing spectrum dimensions."""
        return f"Spectrum dimensions (windows, ft, fy, fx): {self.shape}"

    def __getitem__(self, idx):
        """
        Generate on-the-fly spectra for subwindows for the requested index or slice.
        """

        # Use the sliding_window_func to generate the windows
        if isinstance(idx, slice):
            windows_sel = self.windows[idx]
        else:
            windows_sel = self.windows[idx : idx + 1]
        # now apply spectra on windows
        spectrum = spectral.sliding_window_spectrum(
            windows_sel,
            self.time_size,
            self.time_overlap,
        )

        # preprocess
        spectrum = spectral.spectrum_preprocessing(
            spectrum,
            self.kt,
            self.ky,
            self.kx,
            self.smax*3,
            spectrum_threshold=self.threshold
        )
        if isinstance(idx, slice):
            return spectrum
        elif isinstance(idx, Union[int, np.integer]):
            return spectrum[0]
        else:
            raise TypeError(f"Invalid index type: {type(idx)}. Only int or slice is supported.")

    def __len__(self):
        """
        Return the total number of windows of spectram
        """
        return len(self.windows)

    @property
    def shape(self):
        return len(self), self.kt.size, self.ky.size, self.kx.size



import numba as nb
import numpy as np


@nb.njit(cache=True)
def fft_x(x):
    return np.fft.fft(x, axis=-1)


@nb.njit(cache=True)
def fft_y(x):
    return np.fft.fft(x, axis=-2)


@nb.njit(cache=True)
def ifft_(x):
    return np.fft.ifft(x, axis=0)


@nb.njit(cache=True)
def fftshift_(x):
    return np.fft.fftshift(x)


def wave_numbers(
    window_dims: tuple,
    res: float,
    fps: float
):
    """
    get t, y, x wave numbers

    Parameters
    ----------
    windows : np.ndarray
        time x Y x X windows with intensities
    res : float
        resolution in xy direction
    fps : float
        frames per second

    Returns
    -------
    kt, ky, kx: np.ndarray
        wave numbers of time, y and x

    """
    ks = 2 * np.pi / res  # this assumes the resolution is the same in x and
    # y-direction: TODO make variable for both directions
    kts = 2* np.pi * fps # change frequency units to rad/s
    dkt = kts / window_dims[-3]
    dky = ks / window_dims[-2]
    dkx = ks / window_dims[-1]
    # omega wave numbers (time dim)
    kt = np.arange(0, kts, dkt)
    kt = kt[0:np.int64(np.ceil(len(kt) / 2))]

    # determine wave numbers in x-direction
    kx = np.arange(0, ks, dkx)
    # kx = 0:dkx: (ks - dkx)
    ky = np.arange(0, ks, dky)

    # apply fftshift on determined wave numbers
    kx = np.fft.fftshift(kx)
    ky = np.fft.fftshift(ky)
    idx_x0 = np.where(kx == 0)[0][0]
    kx[0: idx_x0] = kx[0:idx_x0] - kx[idx_x0 - 1] - dkx
    idx_y0 = np.where(ky == 0)[0][0]
    ky[0: idx_y0] = ky[0:idx_y0] - ky[idx_y0 - 1] - dky

    return kt, ky, kx


@nb.njit(parallel=True, cache=True, nogil=True)
def _numba_fourier_transform(
    windows: np.ndarray
) -> np.ndarray:
    """
    Perform 3D spectral analysis with numba jitted code.

    Parameters
    ----------
    windows : np.ndarray
        time x Y x X windows with intensities

    Returns
    -------
    spectrum : np.ndarray
        3D power spectrum of 3D fourier transform

    """
    spectrum_2d = np.empty(windows.shape, dtype=np.complex128)
    for n in nb.prange(windows.shape[0]):
        spectrum_2d[n] = fftshift_(
        fft_y(
            fft_x(windows[n])
        )
    )
    spectrum_3d = ifft_(spectrum_2d)
    # return spectrum_3d
    power = np.abs(spectrum_3d) ** 2
    # abbreviate to positive omega
    return power[:int(np.ceil(len(power)/2))]


@nb.njit(parallel=True, cache=True, nogil=True)
def _numba_fourier_transform_multi(
    imgs: np.ndarray
) -> np.ndarray:
    """
    Perform 3D spectral analysis for multiple windows at once with numba jitted code.

    Parameters
    ----------
    imgs : np.ndarray
        n x time x Y x X windows with intensities

    Returns
    -------
    spectra : np.ndarray
        n x 3D power spectrum of 3D fourier transform of all imgs

    """
    spectra = np.empty((imgs.shape[0], int(np.ceil(imgs.shape[1]/2)), imgs.shape[2], imgs.shape[3]), dtype=np.float64)
    for m in nb.prange(imgs.shape[0]):
        spectrum_3d = np.empty(imgs[m].shape, dtype=np.complex128)
        for n in nb.prange(imgs.shape[1]):
            spectrum_3d[n] = fftshift_(
                fft_y(
                    fft_x(imgs[m, n])
                )
            )
        for y in nb.prange(spectrum_3d.shape[1]):
            for x in nb.prange(spectrum_3d.shape[2]):
                spectrum_3d[:, y, x] = ifft_(spectrum_3d[:, y, x])
        # return spectrum_3d
        power = np.abs(spectrum_3d) ** 2
        # abbreviate to positive omega
        spectra[m] = power[:int(np.ceil(len(power)/2))]
    return spectra


def _numpy_fourier_transform(
        windows: np.ndarray,
        norm: bool = False
) -> np.ndarray:
    """
    Pure numpy implementation of 3D spectral analysis

    Parameters
    ----------
    windows : np.ndarray
        time x Y x X windows with intensities
    norm : bool
        normalize spectrum (default: False)

    Returns
    -------
    power : np.ndarray
        3D power spectrum of 3D fourier transform

    """
    spectrum_2d = np.fft.fftshift(
        np.fft.fft(np.fft.fft(windows, axis=-2), axis=-1)
    )
    spectrum_3d = np.fft.ifft(spectrum_2d, axis=0)
    spectrum = np.abs(spectrum_3d) ** 2

    # abbreviate to positive omega
    spectrum = spectrum[:int(np.ceil(len(spectrum)/2))]

    if norm:
        spectrum_norm = spectrum / np.expand_dims(
            spectrum.mean(axis=-1).mean(axis=-1),
            axis=(-1, -2)
        )
        return spectrum_norm
    return spectrum


def spectral_imgs(
    imgs: np.ndarray,
    **kwargs
) -> np.ndarray:
    """
    Perform 3D spectral analysis.

    Parameters
    ----------
    imgs : np.ndarray
        [n * t * Y * X] 4-D array containing image [n] sequences [t], split in
        subwindows of Y * X pixels
    engine : str, optional
        "numpy" or "numba", compute method to use, typically numba (default) is
        a lot faster. Numpy function is easier to read.
    kwargs : dict with additional keyword arguments for processing

    Returns
    -------
    spectra : np.ndarray
        wave spectra for all image window sequences

    """
    n, t, y, x = imgs.shape
    t_out = int(np.ceil(t / 2))  # Reduced time dimension
    spectra =  _numba_fourier_transform_multi(imgs, **kwargs)
    return spectra


def sliding_window_spectrum(
    imgs: np.ndarray,
    win_t: int,
    overlap: int,
    **kwargs
) -> np.ndarray:
    """
    Splits the video into shorter segments and calculates the average 3D spectrum.

    Parameters
    ----------
    imgs : np.ndarray
        [n * t * Y * X] 4-D array containing image [n] sequences [t], split in 
        subwindows of Y * X pixels
    win_t : int
        number of frames per segment
    overlap : int
        overlap (frames)
    kwargs : dict with additional keyword arguments for processing

    Returns
    -------
    spectra : np.ndarray
        average wave spectra for all image window sequences

    """
    
    # Check for division by zero
    if win_t == overlap:
        raise ValueError("win_t and overlap should not be equal.")
    
    # number of segments
    num_segments = (imgs.shape[1] - win_t) // (win_t - overlap)+ 1
    
    # sum of individual segments
    spectrum_sum = sum(
        spectral_imgs(
            imgs[:, segment_t0:(segment_t0 + win_t), :, :],
            **kwargs
        ) for segment_t0 in range(0, imgs.shape[1] - win_t + 1, win_t - overlap)
    )
    
    # renormalisation
    spectra = spectrum_sum / num_segments
    return spectra


def spectrum_preprocessing(
        measured_spectrum: np.ndarray,
        kt: np.ndarray,
        ky: np.ndarray,
        kx: np.ndarray,
        velocity_threshold: float,
        spectrum_threshold: float = 1
) -> np.ndarray:
    """
    pre-processing of the measured spectrum to improve convergence of the optimisation

    Parameters
    ----------
    measured_spectrum : np.ndarray
        measured, averaged, and normalised 3D power spectrum calculated with spectral.py
        dimensions [wi, kti, kyi, kx]

    kt : np.ndarray
        radian frequency vector (rad/s)

    ky : np.ndarray
        y-wavenumber vector (rad/m)

    kx : np.ndarray
        x-wavenumber vector (rad/m)

    velocity_threshold : float
        maximum threshold velocity for spectrum filtering (m/s).

    spectrum_threshold : float, optional
        threshold parameter for spectrum filtering (default 1.0).
        the spectrum with amplitude < threshold_preprocessing * mean(measured_spectrum) is filtered out.
        threshold_preprocessing < 1 yields a more severe filtering but could eliminate part of useful signal.

    Returns
    -------
    preprocessed_spectrum : np.ndarray
        pre-processed and normalised measured 3D spectrum

    """
    # spectrum normalisation: divides the spectrum at each frequency by the average across all wavenumber combinations at the same frequency
    with np.errstate(divide='ignore', invalid='ignore'):
        preprocessed_spectrum = measured_spectrum / np.mean(measured_spectrum, axis=(2, 3), keepdims=True)

    # apply threshold
    threshold = spectrum_threshold * np.mean(preprocessed_spectrum, axis=1, keepdims=True)
    preprocessed_spectrum[preprocessed_spectrum < threshold] = 0

    # set the first slice (frequency=0) to 0
    preprocessed_spectrum[:, 0, :, :] = 0

    kt_threshold = dispersion_threshold(ky, kx, velocity_threshold)

    # set all frequencies higher than the threshold frequency to 0
    kt_reshaped = kt[:, np.newaxis, np.newaxis]  # reshape kt to be broadcastable
    kt_threshold_bc = np.broadcast_to(kt_threshold, (kt.shape[0], kt_threshold.shape[1], kt_threshold.shape[
        2]))  # broadcast kt_threshold to match the dimensions of kt
    kt_bc = np.broadcast_to(kt_reshaped, kt_threshold_bc.shape)  # broadcast kt to match the dimensions of kt_threshold
    mask = np.where(kt_bc <= kt_threshold_bc, 1, 0)  # create mask
    mask = np.expand_dims(mask, axis=0)

    preprocessed_spectrum = preprocessed_spectrum * mask  # apply mask

    # remove NaNs
    preprocessed_spectrum = np.nan_to_num(preprocessed_spectrum)
    return preprocessed_spectrum


def dispersion_threshold(
        ky,
        kx,
        velocity_threshold
) -> np.ndarray:
    """
    Calculate the frequency corresponding to the threshold velocity

    Parameters
    ----------
    ky: np.ndarray
        wavenumber array along the direction y

    kx: np.ndarray
        wavenumber array along the direction x

    velocity_threshold : float
        threshold_velocity (m/s)

    Returns
    -------
    kt_threshold : np.ndarray
        1 x N_y x N_x: threshold frequency

    """

    # create 2D wavenumber grid
    kx, ky = np.meshgrid(kx, ky)

    # transpose to 1 x N_y x N_x
    ky = np.expand_dims(ky, axis=0)
    kx = np.expand_dims(kx, axis=0)

    # wavenumber modulus
    k_mod = np.sqrt(ky ** 2 + kx ** 2)

    return k_mod * velocity_threshold

